#ifndef _TDA6651_H
#define _TDA6651_H

#include "Prefix.h"
#include "TC90512.h"

namespace EARTH {
	class TDA6651 {
	public:
		TDA6651(TC90512 *tc90512) : mTC90512(tc90512)
		{
			_ASSERTE(mTC90512);

			mDataIsInvalid = true;
		}

		~TDA6651()
		{
		}

		enum Mode {
			Mode_Normal,
			Mode_Search
		};

		void Reset()
		{
			mDataIsInvalid = true;
		}

		void Write(uint frequency, Mode mode)
		{
			ushort n = N(frequency);
			_ASSERTE((n >> 15) == 0);

			byte cp;
			byte bs;
			CP_BS(frequency, &cp, &bs);

			byte atc = (mode == Mode_Search) ? 1 : 0;

			byte data[COUNT_DATA];

			data[0] = n >> 8;
			data[1] = static_cast<byte>(n);
			data[2] = BIT_2(1,1) << 6 | BIT_3(0,0,1) << 0;
			data[3] = cp << 5 | bs << 0;
			data[4] = BIT_2(1,0) << 6 | atc << 3 | BIT_3(0,0,0) << 0;

			Write(data);
		}

		void Read(bool *locked)
		{
			byte data;
			mTC90512->ReadThrough(TC90512::Through_T, ADDRESS, &data, sizeof(data));

			_ASSERTE(locked);
			*locked = BIT_SHIFT_MASK(data, 6, 1) != 0;
		}

	protected:
		enum {
			ADDRESS = BIT_7(1,1,0,0,0,0,1),
			COUNT_DATA = 5
		};

		TC90512 *const mTC90512;
		bool mDataIsInvalid;
		byte mData[COUNT_DATA];

		ushort N(uint frequency) const
		{
			return static_cast<ushort>(frequency + 7*57);
		}

		void CP_BS(uint frequency, byte *cp, byte *bs) const
		{
			struct Item {
				uint Frequency;
				byte CP;
				byte BS;
			};

			static const Item TABLE[] = {
				{  90, 4, 0 },

				{ 140, 4, 1 },
				{ 170, 5, 1 },

				{ 220, 3, 2 },
				{ 330, 5, 2 },
				{ 402, 7-1, 2 },

				{ 450, 3, 3 },
				{ 550, 4, 3 },
				{ 600, 5, 3 },
				{ 700, 6, 3 },
				{ 770, 7, 3 }
			};

			const Item *item = NULL;
			for (uint i=0; i<sizeof(TABLE)/sizeof(*TABLE); i++) {
				const Item *ptr = &TABLE[i];
				if (frequency < ptr->Frequency*7) {
					item = ptr;
					break;
				}
			}

			_ASSERTE(item);
			_ASSERTE(cp); *cp = item->CP;
			_ASSERTE(bs); *bs = item->BS ? (BIT_1(1) << (item->BS-1)) : 0;
		}

		void Write(const byte *fullData)
		{
			bool set[COUNT_DATA];

			for (uint i=0; i<sizeof(set)/sizeof(*set); i++) {
				set[i] = mDataIsInvalid || (fullData[i] != mData[i]);
			}

			byte data[COUNT_DATA];
			uint size = 0;

			if (set[0] || set[1]) {
				data[size++] = fullData[0];
				data[size++] = fullData[1];
				set[0] = false;
				set[1] = false;
			}

			if (set[3]) {
				if (set[2]) {
					data[size++] = fullData[2];
					data[size++] = fullData[3];
					set[2] = false;
					set[3] = false;
				} else {
					data[size++] = fullData[4];
					data[size++] = fullData[3];
					set[4] = false;
					set[3] = false;
				}
			}

			if (set[2]) {
				data[size++] = fullData[2];
				set[2] = false;
			}

			if (set[4]) {
				data[size++] = fullData[4];
				set[4] = false;
			}

			for (uint i=0; i<sizeof(set)/sizeof(*set); i++) {
				_ASSERTE(set[i] == false);
			}

			mTC90512->WriteThrough(TC90512::Through_T, ADDRESS, data, size);

			for (uint i=0; i<sizeof(mData)/sizeof(*mData); i++) {
				mData[i] = fullData[i];
			}
			mDataIsInvalid = false;
		}
	};
}

#endif
